% this function try to solve the Joint matrix completion model 3, i.e.
% X_i=C+A_i+E_i
% C_hat is initialized as the mean of all training samples;
% and it is modified from exact PCA model
% =========================================================================
% inputs:
% X:            input signals, ?R^(m*n)*N, where N is total number of signals;
% N:            number of signals;
% lambda:        the weight for l1-regularizer term;
% tol:          tolerance for stopping criterion;
% maxIters:     maximal number of iterations;
% limem:        if limen==1, the temporary variable will be cleared to save
% memory;
% =========================================================================
% A:            the common part, which is low rank matrix, ?R^(m*n);
% E:            the innovation part, which is sparse, ?R^(m*n)*N;
% output:       some other output;
% =========================================================================
% version
% 3:            modify the parameters mu and Y, by incorporating C_hat;
% 2:            use weighted average to update;
% 1:            the first version of the code, some parameters tune up;
% =========================================================================
function [C_hat, A_hat, E_hat, output]=JointMC3(X, N, lambda, tol, maxIter, limem, weight, debug, tau, rho)
%% ========================================================================
% check the paramets
disp('check the paramets');
% data size
[m n]=size(X);
n=n/N;
disp([m n]);
% setup lambda
if nargin < 3
    lambda = 1 / sqrt(m);
end
disp(lambda);
% tolerance
if nargin < 4
    tol = 1e-7;
elseif tol == -1
    tol = 1e-7;
end
disp(tol);
% max number of iterations
if nargin < 5
    maxIter = 1000;
elseif maxIter == -1
    maxIter = 1000;
end
disp(maxIter);
if nargin < 6
    limem =0;
end
disp(limem);
% united weight
if nargin < 7
    weight=1;
end
% bebug mode
if nargin < 8
    debug=0;
end
% step size
if nargin < 9
    tau=1.25;
end
if nargin < 9
    rho=1.5;
end
%% ========================================================================
% % preprocess the data
% disp('preprocess the data');
% % zero mean
% tmpx=reshape(X, m*n, N);
% avg=mean(tmpx);
% tmpx=tmpx-repmat(avg, m*n, 1);
% % unit l2 norm
% mag=sqrt(sum(tmpx.*tmpx));
% tmpx=tmpx./repmat(mag, m*n, 1);
% % convert to X
% X=reshape(tmpx, m, n*N);
% clear tmpx;
%% ========================================================================
% initialization
disp('initialization');
% parameters
dnorm = norm(X, 'fro');
tolProj = 1e-6 * dnorm;
total_svd = 0;
% rho = 1.5;          % this one can be tuned
% tau=1.25;
% rho = 6;          % this one can be tuned
% variable
% C_hat = zeros( m, n);
A_hat = zeros( m, n*N);
E_hat = zeros( m, n*N);
iter = 0;
converged = false;
stopCriterion = 1;
sv = 5*ones(1, N);
svp = sv;
% =========================================================================
% common is initialize as the mean of the frames
C_hat=reshape(mean(reshape(X, m*n, N), 2), m, n);
% Y is the weight for lagarian multiplier
Y=zeros(size(X));
mu=zeros(1, N);
if weight
    for i=1: N
        % for each Y
        tmpy=X(:, (i-1)*n+1: i*n)-C_hat;
        norm_two = lansvd(tmpy, 1, 'L');
        norm_inf = norm( tmpy(:), inf) / lambda;
        dual_norm = max(norm_two, norm_inf);
        tmpy = tmpy / dual_norm;
        Y(:, (i-1)*n+1: i*n)=tmpy;
        tmpmu = tau/norm_two; % this one can be tuned
        % tmpmu = .5/norm_two; % this one can be tuned
        mu(i)=tmpmu;
    end
    clear tmpy tmpmu i norm_two norm_inf dual_norm;
else
    Y=X-repmat(C_hat, 1, N);
    norm_two = lansvd(Y, 1, 'L');
    norm_inf = norm( Y, inf) / lambda;
    dual_norm = max(norm_two, norm_inf);
    Y=Y/dual_norm;
    mu = repmat(tau/norm_two, 1, N); % this one can be tuned
    clear norm_two norm_inf dual_norm;
end
%% ========================================================================
% main loop starts here
disp('% main loop starts here');
while ~converged
    % outer loop
    iter = iter + 1;
    % solve the primal problem by alternative projection
    primal_converged = false;
    primal_iter = 0;
    sv = sv + round(n * 0.1);
    %% ====================================================================
    while primal_converged == false
        %% =================================================================
        % part 1
        % part 1: solve E when A is fixed
        % L(E) = lambda * |E|_1 + <Y,D-A-E> + mu/2 *|D-A-E|_F^2;
        if debug
            disp('part 1: sparse part');
        end
        temp_E=zeros(m, n*N);
        temp_A=zeros(m, n*N);
        temp_C=zeros(m, n);
        for i=1:N
            tmpx=X(:, (i-1)*n+1: i*n);
            tmpy=Y(:, (i-1)*n+1: i*n);
            tmpa=A_hat(:, (i-1)*n+1: i*n);
            temp_T = tmpx - C_hat -tmpa + (1/mu(i))*tmpy;
            tmpe=max( temp_T - lambda/mu(i),0) + min( temp_T + lambda/mu(i),0);
            temp_E(:, (i-1)*n+1: i*n)=tmpe;
        end
        if limem
            clear i tmpx tmpy tmpe tmpa temp_T;
        end
        %% ================================================================
        % part 2: low rank part
        % A_i=argmin_(A_i ) |A_i |_*+?|(X_i-C)-A_i |_F+<Y,X_i-C-A_i>
        if debug
            disp('part 2: low rank part');
        end
        for i=1:N
            tmpx=X(:, (i-1)*n+1: i*n);
            tmpy=Y(:, (i-1)*n+1: i*n);
            tmpe=temp_E(:, (i-1)*n+1: i*n);
            if choosvd(n, sv(i)) == 1
                [U S V] = lansvd(tmpx - C_hat-tmpe + (1/mu(i))*tmpy, sv(i), 'L');
            else
                [U S V] = svd(tmpx - C_hat-tmpe + (1/mu(i))*tmpy, 'econ');
            end
            diagS = diag(S);
            svp(i) = length(find(diagS > 1/mu(i)));
            if svp(i) < sv(i)
                sv(i) = min(svp(i) + 1, n);
            else
                sv(i) = min(svp(i) + round(0.05*n), n);
            end
            tmpa = U(:,1:svp(i))*diag(diagS(1:svp(i))-1/mu(i))*V(:,1:svp(i))';    
            temp_A(:, (i-1)*n+1: i*n)=tmpa;
        end
        if limem
            clear i tmpx tmpy tmpe tmpE tmpa U S V;
        end
        %% ================================================================
        % part 3: common part
        % C=argmin_C ?_i|X_i-C-A_i |_F +<Y,X_i-C-A_i>
        % solution is C=1/n ?_i?X_i^*+Y?
        if debug
            disp('part 3: common part');
        end
        for i=1: N
            tmpx=X(:, (i-1)*n+1: i*n)-temp_A(:, (i-1)*n+1: i*n)-...
                temp_E(:, (i-1)*n+1: i*n);
% =========================================================================
            % do we need it?
            tmpy=Y(:, (i-1)*n+1: i*n);
            tmpc=mu(i)*tmpx+tmpy;
%             tmpc=tmpx+(1/mu(i))*tmpy;
% =========================================================================
            temp_C(:, (i-1)*n+1: i*n)=tmpc;
        end
        % wighted average the A
        tmpc=reshape(sum(reshape(temp_C, m*n, N), 2), m, n)/sum(mu);
        if limem
            clear i tmpx tmpy temp_C;
        end
        %% ================================================================
        % some other stuff
        % inner loop
        if debug
            disp('part3: prepare the result');
        end
        % detect the convergence
        if norm(A_hat - temp_A, 'fro') < tolProj && ...
                norm(C_hat - tmpc, 'fro') < tolProj && ...
                norm(E_hat-temp_E, 'fro') <tolProj
            primal_converged = true;
        end
        C_hat = tmpc;
        A_hat= temp_A;
        E_hat=temp_E;
        primal_iter = primal_iter + 1;
        total_svd = total_svd + 1;
        if limem
            clear temp_A tmpc;
        end
        fprintf('%dth inner loop: %d svd operations used\n', primal_iter, total_svd);
    end
    %% ====================================================================
    % complete inner loop
    % update Y
    Z=X-repmat(C_hat, 1, N)-A_hat-E_hat;
    for i=1:N
        tmpy=Y(:, (i-1)*n+1: i*n);
        tmpz=Z(:, (i-1)*n+1: i*n);
        tmpy=tmpy+tmpz*mu(i);
        Y(:, (i-1)*n+1: i*n)=tmpy;
    end
    if limem
        clear tmpy tmpz i;
    end
    % some modification here
    % update mu: mu=tau/|X_i-C_hat|_2
%     if weight
%         for i=1: N
%             % for each Y
%             tmpy=X(:, (i-1)*n+1: i*n)-C_hat;
%             norm_two = lansvd(tmpy, 1, 'L');
%             tmpmu = tau/norm_two; % this one can be tuned
%             % tmpmu = .5/norm_two; % this one can be tuned
%             mu(i)=tmpmu;
%         end
%         clear tmpy tmpmu i norm_two;
%     else
%         tmpy=X-repmat(C_hat, 1, N);
%         norm_two = lansvd(Y, 1, 'L');
%         mu = repmat(tau/norm_two, 1, N); % this one can be tuned
%         clear norm_two tmpy;
%     end
    mu=rho*mu;
    %% ====================================================================
    % stop Criterion    
    stopCriterion = norm(Z, 'fro') / dnorm;
    if stopCriterion < tol
        converged = true;
    end    
    
    disp(['Iteration' num2str(iter) ' #svd ' num2str(total_svd)...
        ' stopCriterion ' num2str(stopCriterion)]);
    disp([' r(A) ' num2str(svp)]);
    disp([' mu ' num2str(mu)]);
    
    if ~converged && iter >= maxIter
        disp('Maximum iterations reached') ;
        converged = 1 ;       
    end
    clear Z;
    save('data.mat');
    % ====================================================================
    % plot
    if debug
        % plot
        figure('name', ['Iteration' num2str(iter) ' #svd ' ...
            num2str(total_svd) ' stopCriterion ' num2str(stopCriterion)]);
        % =================================================================
        % original images
        subplot(221);
        img=zeros(192*8, 168*8);
        for irow=1:8
            img((irow-1)*192+1: irow*192, :)=X(:, (irow-1)*168*8+1: irow*168*8);
        end
        imagesc(img); axis image; colormap gray;
        title('original images');
        % =================================================================
        % 'common part'
        clear img irow;
        subplot(222);
        imagesc(C_hat);colormap gray;axis image;
        title('common part')
        % =================================================================
        % innovation part: low rank matrix
        subplot(223);
        img=zeros(192*8, 168*8);
        for irow=1:8
            img((irow-1)*192+1: irow*192, :)=A_hat(:, (irow-1)*168*8+1: irow*168*8);
        end
        imagesc(img); axis image; colormap gray;
        title('innovation part: low rank matrix');
        clear img irow;
        % =================================================================
        % sparse
        subplot(224);
        img=zeros(192*8, 168*8);
        for irow=1:8
            img((irow-1)*192+1: irow*192, :)=E_hat(:, (irow-1)*168*8+1: irow*168*8);
        end
        imagesc(img); axis image; colormap gray;
        title('innovation part: sparse matrix');
        clear img irow;
        % =================================================================
        % others
        waitforbuttonpress();
        close;
    end
end
output.svp=svp;
output.iter=iter;
output.stopCriterion=stopCriterion;
output.mu=mu;
output.X=X;
