% this script implements the SRC method, when there are missing training
% entries. We use partition method, the best combination is OMP, normalized
% error and no normalization to each partition of dictionary.
% parameters:
%   dataset:        dataset you want to test, 'AR' or 'YaleB';
%   path:           the path of the dataset;
%   n_People:       number of the subjects in the dataset;
%   n_Condition:    how many images per subjects;
%   height:         the height of image (for resize);
%   width:          the width of image;
%   stage:          the procedure that you want to process, [abc]_2, where
%       a means the training stage, b testing for testing sets and c for
%       testing of testing set;
%   debug:          the control of debug output. when set, information and
%       variable will be saved;
%   limem:          if set, the function works in memory limited mode;
%   mode:           controls the normalization method, default 2:
%       0   :   do nothing;
%       1   :   unit l2 norm for each image;
%       2   :   unit l2 norm and zero mean for each image;
%       3   :   unit std and zero mean for each image;
%   tol:            the error tolerance for GPSR;
%   maxIter:        the maximal number of iterations for GPSR;
%   tau:            the weight for sparsity of GPSR;
%   ratio:          controls the ratio of missing entries or the sampling 
%       matrix;
% Example
% demoSRC('YaleB', 'F:\MATLAB\YaleBFull\', 38, 32, 48, 42, 3, 0, 1, 2);
% =========================================================================
function demoSRC(dataset, path, n_People, n_Condition, ...
    height, width, stage, debug, limem, mode, ...
    tol, maxIter, tau, ratio)
%% ========================================================================
% preprocessing stage
% the name of the dataset
% the path of the dataset
% number of people in the database
if nargin<4
    disp('usage: runJointMC.exe dataset path n_People n_Condition');
    exit;
end
option.dataset=dataset;
option.path=path;
if isnumeric(n_People)
    option.n_People=n_People;
else
     option.n_People=str2double(n_People);
end
if isnumeric(n_Condition)
    option.n_Condition=n_Condition;
else
     option.n_Condition=str2double(n_Condition);
end
% optional parameters
if nargin<5 || isempty(height) || isempty(width)
    option.height=0;
    option.width=0;
else
    if isnumeric(height)
        option.height=height;
    else
         option.height=str2double(height);
    end
    if isnumeric(width)
        option.width=width;
    else
         option.width=str2double(width); 
    end
end
% control the process: stage=[abc]2;
% a=1:  run training process
% b=1:  evaluate training data
% c=1:  evaluate testing data
if nargin<7 || isempty(stage)
    option.stage=7;
else
    if isnumeric(stage)
        option.stage=stage;
    else
         option.stage=str2double(stage); 
    end
end
% debug=1: debug mode, all information will be displayed and saved
if nargin<8 || isempty(debug)
    option.debug=0;
else
    if isnumeric(debug)
        option.debug=debug;
    else
         option.debug=str2double(debug); 
    end
end
% limem=1: memory limited mode, temporary variable will be automatically
% cleared. Speed maybe delayed
if nargin<9 || isempty(limem)
    option.limem=1;
else
    if isnumeric(limem)
        option.limem=limem;
    else
         option.limem=str2double(limem); 
    end
end
% normalization method, default 2:
% 0 : do nothing;
% 1 : unit l2 norm for each image;
% 2 : unit l2 norm and zero mean for each image;
% 3 : unit std and zero mean for each image;
if nargin<10 || isempty(mode)
    option.mode=2;
else
    if isnumeric(mode)
        option.mode=mode;
    else
         option.mode=str2double(mode); 
    end
end
% error tolerance (default 1e-6)
if nargin<11 || isempty(tol)
    option.tol=0.05;
else
    if isnumeric(tol)
        option.tol=tol;
    else
         option.tol=str2double(tol); 
    end
end
% max number of iterations (default 30)
if nargin<12 || isempty(maxIter)
    option.maxIter=1000;
else
    if isnumeric(maxIter)
        option.maxIter=maxIter;
    else
         option.maxIter=str2double(maxIter); 
    end
end
% tau (default 1.25)
if nargin<13 || isempty(tau)
    option.tau=0.009;
else
    if isnumeric(tau)
        option.tau=tau;
    else
         option.tau=str2double(tau); 
    end
end
% gross to control whether pack the E_hat into G_hat
if nargin<14 || isempty(ratio)
    option.ratio=0;
else
    if ~isscalar(ratio)
        % load in omega
        omega=ratio;
        option.ratio=sum(ratio(:))/numel(ratio);
    elseif isnumeric(ratio)
        option.ratio=ratio;
    else
        option.ratio=str2double(ratio); 
    end
end
% show the parameter
disp(option);
%% ========================================================================
% training stage
if bitand(option.stage, 1)
    % some variable
    disp('training dictionary');
    if ~exist('omega') %#ok<EXIST>
        omega=GenerateSample(option.n_People, option.n_Condition, option.ratio);
    end
    % calculate the condition number, i.e. the ratio of missing entries
    fprintf(1, 'total:\t%.2f\nsubject:\t%.2f\ncondition:\t%.2f', ...
        sum(omega(:))/(option.n_People*option.n_Condition), ...
        min(sum(omega, 2))/option.n_Condition, ...
        min(sum(omega, 1))/option.n_People);
    % calculating the boundary
    bound=zeros(1, option.n_People+1);
    for i=1: option.n_People
        bound(i+1)=bound(i)+sum(omega(i, :));
    end
    % the random matrix
    cd(path);
    dict=TrainSRC(omega, option);
    % wired that, nomalization gets a worse result
%     dict=BlockDict(dict, option);
    % to gain robustness
    save([dataset '-SRC-train.mat'], 'dict', 'omega', 'bound', 'option');
else
    disp('load dictionary');
    load([dataset '-SRC-train.mat'], 'dict', 'bound', 'omega');
end
%% ========================================================================
% start the testing
% for each test image, which should be stored under path\test\
if bitand(option.stage, 2)
    % for GPSR
    f1=fopen([dataset '-test.txt'], 'w');
    cur_dir=pwd();
    cd([path 'test/']);
    if strcmpi(dataset, 'PIE') || strcmpi(dataset, 'Multi_PIE')
        dirlist=dir('*.jpg');
    else
        dirlist=dir('*.bmp');
    end
    idx=zeros(1, length(dirlist));
    clear output;
    try
        for i=1:length(dirlist)
            % for block based
            [idx(i), output(i)]=BlockError(dirlist(i).name, dict, bound, option);
%             [idx(i), output(i)]=TestSRC(dirlist(i).name, dict, bound, option);
           fprintf(f1, '%s result is %03d\n', dirlist(i).name, idx(i));
        end
        cd(cur_dir);
        save([dataset datestr(now, '-yy-mm-dd-HH-MM-SS-') 'SRC-result.mat'],...
            'idx', 'output', 'option', 'bound', 'omega');
    catch ME
        getReport(ME)
        cd(cur_dir);
        save([dataset '-SRC-exception.mat']);
    end
    fclose(f1);
end
end

%% the training function
function dict=TrainSRC(omega, option)
    if strcmpi(option.dataset, 'PIE') || strcmpi(option.dataset, 'Multi_PIE')
        dirlist=dir('*.jpg');
    else
        dirlist=dir('*.bmp');
    end
    % build the dictionary
    dict=zeros(option.width*option.height, sum(omega(:)));
    k=1;
    for i=1: option.n_People
        for j=1: option.n_Condition
            if omega(i, j)
                img=imread(dirlist((i-1)*option.n_Condition+j).name);
                if size(img, 3)~=1
                    img=rgb2gray(img);
                end
                img=imresize(img, [option.height option.width]);
                img=double(img);
                % normalization needed
                switch option.mode
                    case 0
                        % do nothing
                    case 1
                        % unit l2-norm for each image
                        img=img/norm(img, 'fro');
                    case 2
                        % zero mean and unit l2-norm for each image
                        img=img-mean(img(:));
                        img=img/norm(img, 'fro');
                    case 3
                        % zero mean and unit std for eac dimension
                        img=img/255;
                        img=img-mean(img(:));
                    case 4
                        % zerp mean
                        img=img-mean(img(:));
                end
                dict(:, k)=img(:);
                k=k+1;
            end
        end
    end
end

%% for testing function
function [idx, output]=TestSRC(filename, dict, bound, option)
    % read in the image
    img=imread(filename);
    if size(img, 3)~=1
        img=rgb2gray(img);
    end
    img=imresize(img, [option.height option.width]);
    img=double(img);
    % normalization needed
    switch option.mode
        case 0
            % do nothing
        case 1
            % unit l2-norm for each image
            img=img/norm(img, 'fro');
        case 2
            % zero mean and unit l2-norm for each image
            img=img-mean(img(:));
            img=img/norm(img, 'fro');
        case 3
            % zero mean and unit std for eac dimension
            img=img/255;
            img=img-mean(img(:));
        case 4
            % zerp mean
            img=img-mean(img(:));
    end
    img=img(:);
    % testing
%     output.coef=OMP(dict, img, 2*option.n_Condition);
    x0=dict'*img;
    first_tau_factor = 0.8*(max (max((abs(x0)./option.tau))));
    output.coef=GPSR_BB(img, dict, option.tau,...
         'Debias',0,...
         'Continuation',1,...
         'ContinuationSteps',5,...
         'FirstTauFactor',first_tau_factor,...
         'Monotone',1,...
         'Initialization',x0,...
         'StopCriterion',3,...
         'ToleranceA',option.tol,...
         'Verbose',0);
    % so we try E_hat
    output.err=zeros(1, option.n_People);
    for j=1: option.n_People
        output.err(j)=norm(img-dict(:, bound(j)+1: bound(j+1))*...
            output.coef(bound(j)+1: bound(j+1)), 2);
    end
    [val, idx]=min(output.err); %#ok<ASGLU>
end

%% normalize the dictionary for partion based method
function dict=BlockDict(dict, option)
    if strcmp(option.dataset, 'AR')
        for i=[1 10 19 28 38]
            for j=[1 8 15 22 27]
                mask=zeros(option.height, option.width);
                mask(i: i+17, j: j+13)=1;
                for l=1: size(dict, 2)
                    dict(mask==1, l)=dict(mask==1, l)./norm(dict(mask==1, l));
                end
            end
        end
    elseif strcmp(option.dataset, 'YaleB')
        for i=[1 9 17 25 33]
            for j=[1 8 15 22 29]
                mask=zeros(option.height, option.width);
                mask(i: i+15, j: j+13)=1;
                for l=1: size(dict, 2)
                    dict(mask==1, l)=dict(mask==1, l)./norm(dict(mask==1, l));
                end
            end
        end
    elseif strcmpi(option.dataset, 'PIE')
        for i=[1 9 18 26 34]
            for j=[1 6 13 18 24]
                mask=zeros(option.height, option.width);
                mask(i: i+16, j: j+11)=1;
                for l=1: size(dict, 2)
                    dict(mask==1, l)=dict(mask==1, l)./norm(dict(mask==1, l));
                end
            end
        end
    end
end

%% for partition based
function [idx, output]=BlockError(filename, dict, bound, option)
    %% read in the image
    img=imread(filename);
    if size(img, 3)~=1
        img=rgb2gray(img);
    end
    img=imresize(img, [option.height option.width]);
    img=double(img);
    % normalization needed
    switch option.mode
        case 0
            % do nothing
        case 1
            % unit l2-norm for each image
            img=img/norm(img, 'fro');
        case 2
            % zero mean and unit l2-norm for each image
            img=img-mean(img(:));
            img=img/norm(img, 'fro');
        case 3
            % zero mean and unit std for eac dimension
            img=img/255;
            img=img-mean(img(:));
        case 4
            % zerp mean
            img=img-mean(img(:));
    end
    img=img(:);
    %% Variable
    err=zeros(option.n_People, 25);
    l=1;
    %% test
    if strcmp(option.dataset, 'AR')
        for i=[1 10 19 28 38]
            for j=[1 8 15 22 27]
                mask=zeros(option.height, option.width);
                mask(i: i+17, j: j+13)=1;
                mask=(mask(:)==1);
                D=reshape(dict(repmat(mask, 1, size(dict, 2))), [], size(dict, 2));
                coef=OMP(D, img(mask), 2*option.n_Condition);
                parfor k=1: option.n_People
                    err(k, l)=norm(img(mask)-D(:, bound(k)+1: bound(k+1))*...
                        coef(bound(k)+1: bound(k+1)), 2);
                end
                l=l+1;
            end
        end
    elseif strcmp(option.dataset, 'YaleB')
        for i=[1 9 17 25 33]
            for j=[1 8 15 22 29]
                mask=zeros(option.height, option.width);
                mask(i: i+15, j: j+13)=1;
                mask=(mask(:)==1);
                D=reshape(dict(repmat(mask, 1, size(dict, 2))), [], size(dict, 2));
                coef=OMP(D, img(mask), 2*option.n_Condition);
                parfor k=1: option.n_People
                    err(k, l)=norm(img(mask)-D(:, bound(k)+1: bound(k+1))*...
                        coef(bound(k)+1: bound(k+1)), 2);
                end
                l=l+1;
            end
        end
    elseif strcmp(option.dataset, 'PIE')
        for i=[1 9 18 26 34]
            for j=[1 6 13 18 24]
                mask=zeros(option.height, option.width);
                mask(i: i+16, j: j+11)=1;
                mask=(mask(:)==1);
                D=reshape(dict(repmat(mask, 1, size(dict, 2))), [], size(dict, 2));
                coef=OMP(D, img(mask), 2*option.n_Condition);
                parfor k=1: option.n_People
                    err(k, l)=norm(img(mask)-D(:, bound(k)+1: bound(k+1))*...
                        coef(bound(k)+1: bound(k+1)), 2);
                end
                l=l+1;
            end
        end
    end
    %% summarize
    output.err=err;
    % we need to normalize the error
    [val, idx]=min(output.err); %#ok<NASGU>
    err=repmat(val, option.n_People, 1)./output.err;
    [val, idx]=max(sum(err, 2)); %#ok<ASGLU>
    [val, output.idx1]=min(sum(output.err, 2));
    % by mmajority
    [val, tmp]=min(output.err);
    tmp=tabulate(tmp);
    [val, j]=max(tmp(:, 2));
    output.idx2=tmp(j, 1);
end